from Prophet.Legacy import App as _App, ConsoleApp as _ConsoleApp, X11App as _X11App
from Keywords import Keyword as Kw, Set as KwS
from Prophet.Categories import *


class BasiliskII(_App, _X11App):
    name = "Basilisk II"
    comment = "Apple Mac emulator"
    keywords = KwS(Emulator)


class bochs(_App, _ConsoleApp):
    name = "Bochs"
    comment = "x86 architecture emulator"
    keywords = KwS(Emulator)


class dosemu(_App, _ConsoleApp):
    name = "DOSemu"
    comment = "MS-DOS emulator"
    keywords = KwS(Emulator)


class xdosemu(_App, _X11App):
    name = "XDOSemu"
    comment = "MS-DOS emulator, X version"
    keywords = KwS(Emulator)

# TODO : Frodo


class fuse(_App, _X11App):
    name = "Fuse"
    comment = "ZX-Spectrum emulator"
    keywords = KwS(Emulator)


class generator(_App, _X11App):
    name = "Generator"
    comment = "Sega Megadrive emulator"
    keywords = KwS(Emulator)


class snes9x(_App, _X11App):
    name = "Snes9x"
    comment = "Super Nintendo emulator"
    keywords = KwS(Emulator)

# TODO : Vice


class vmware(_App, _X11App):  # FIXME : may reside in its own prefix
    name = "VMWare"
    comment = "x86 virtualizator"
    keywords = KwS(Emulator)

# NOTE : the following two are not interactive

# class wine(_App, _X11App) :
#    name = "Wine"
#    comment = "win32 emulator"
#    keywords = KwS(Emulator)

# class winex(_App, _X11App) :
#    name = "WineX"
#    comment = "win32 emulator"
#    keywords = KwS(Emulator)

# TODO : xmame/xmess

# TODO : vmware, xen


class xz80(_App, _X11App):
    name = "XZ80"
    comment = "ZX-Spectrum emulator"
    keywords = KwS(Emulator)


class xzx(_App, _X11App):
    name = "XZ80"
    comment = "ZX-Spectrum emulator"
    keywords = KwS(Emulator)


class zsnes(_App, _X11App):
    name = "ZSnes"
    comment = "Super Nintendo emulator"
    keywords = KwS(Emulator)



import MenuMaker

from MenuMaker import indent, writeFullMenu


def _map(x):
    for d, s in (("[", "("), ("]", ")")):
        x = x.replace(s, d)
    return x


class Sep(object):

    def emit(self, level):
        return ['%s[separator]' % indent(level)]


class App(object):

    def emit(self, level):
        cmd = self.app.execmd
        if self.app.terminal:
            cmd = MenuMaker.terminal.runCmd(cmd)
        return [
            '%s[exec] (%s) {%s}' %
            (indent(level),
             _map(
                self.app.name),
                cmd)]


class Menu(object):
    tag = "submenu"

    def emit(self, level):
        menu = ['%s[%s] (%s)' % (indent(level), self.tag, _map(self.name))]
        for x in self:
            menu += x.emit(level + 1)
        menu.append('%s[end]' % indent(level))
        return menu


class Root(object):
    tag = "begin"

    def emit(self, level):
        if writeFullMenu:
            return super(Root, self).emit(level)
        else:
            menu = []
            for x in self:
                menu += x.emit(level)
            return menu


class X(MenuMaker.Entry):

    def __init__(self, name, tag, arg=None):
        super(X, self).__init__()
        self.align = MenuMaker.Entry.StickBottom
        self.name = name
        self.tag = tag
        self.arg = arg

    def emit(self, level):
        x = '%s[%s] (%s)' % (indent(level), self.tag, _map(self.name))
        if self.arg:
            x += ' {%s}' % self.arg
        return [x]



import sys
import types
import os
import os.path
import fnmatch
import re
import codecs
import Prophet

from Keywords import Keyword as Kw, Set as KwS
from Paths import ValidPathSet as PS
from Prophet.Categories import *
from Prophet import msg, verbose


class DesktopSet(PS):

    def adopt(self, path):
        path = super(DesktopSet, self).adopt(path)
        for x in os.walk(path):
            if len(fnmatch.filter(x[2], "*.desktop")):
                return path

        raise ValueError


_genMap = (
    ("gnome", GNOME),
    ("kde", KDE),
    ("devel*", Development),
    ("edit*", TextEditor),
    ("educ*", Education),
    ("edut*", Education),
    ("*game*", Game),
    ("arcade", ArcadeGame),
    ("board*", BoardGame),
    ("card*", CardGame),
    ("kids*", KidsGame),
    ("rogue*", AdventureGame),
    ("*tactic*", StrategyGame),
    ("*strat*", StrategyGame),
    ("*toy*", Amusement),
    ("*amus*", Amusement),
    ("*graphic*", Graphics),
    ("*sound*", Audio),
    ("*music*", Audio),
    ("*video*", Video),
    ("*internet*", Network),
    ("*netw*", Network),
    ("*news*", News),
    ("*remote*", Network),
    ("*www*", Network),
    ("*transfer*", FileTransfer),
    ("irc", IRCClient),
    ("*instant*", InstantMessaging),
    ("*messag*", InstantMessaging),
    ("*mail*", Email),
    ("multimedia", AudioVideo),
    ("*docum*", Office),  # ???
    ("*office*", Office),
    ("*graphs*", Chart),
    ("pda", PDA),
    ("*present*", Presentation),
    ("*spread*", Spreadsheet),
    ("*word*proc*", WordProcessor),
    ("*time*manag*", ProjectManagement),
    ("*sett*", Settings),
    ("*syst*", System),
    ("*admin*", System),
    ("accessib*", Accessibility),
    ("archiv*", Archiving),
    ("comm*", Network),
    ("*file*tool*", FileManager),
    ("monitor*", Monitor),
    ("*publish*", Office),
    ("science", Science),
    ("*math*", Math),
    ("*text*tool*", TextEditor),  # ???
    ("*config*", Settings),
    ("*boot*", System),
    ("*init*", System),
    ("*hardw*", HardwareSettings),
    ("packag*", PackageManager),
    ("print*", Settings),
    ("*terminal*", TerminalEmulator),
)

_kdeCfgMap = (
    ("look*feel", Kw("X-KDE-settings-looknfeel")),
    ("*sound*", Kw("X-KDE-settings-sound")),
    ("*syst*", Kw("X-KDE-settings-system")),
    ("accessib*", Kw("X-KDE-settings-accessibility")),
    ("component*", Kw("X-KDE-settings-components")),
    ("netw*", Kw("X-KDE-settings-network")),
    ("secur*", Kw("X-KDE-settings-security")),
    ("*web*", Kw("X-KDE-settings-webbrowsing")),
    ("*desk*", Kw("X-KDE-settings-desktop")),
    ("periph*", Kw("X-KDE-settings-peripherials")),
    ("Hardware", Kw("X-KDE-settings-hardware")),
    ("power*", Kw("X-KDE-settings-power")),
    ("*info*", Kw("X-KDE-information"))
)

# Some ugly Linux distros, notably Mandrake, are known for shipping broken (???!)
# .desktop's which have no Categories section (thx, $!N, mein Freund :) that is the main method for distributing
# the apps throughout the menu thus here's the alternative method for
# guessing the keywords


def _dir2kws(dir):
    """Convert the directory name to the corresponding keyword set, if possible"""
    parts = [x.strip().lower() for x in dir.split("/")]
    if len(
        fnmatch.filter(
            parts,
            "kde")) and len(
            fnmatch.filter(
                parts,
                "*config*") +
        fnmatch.filter(
                parts,
                "*sett*")):
        # Catching KDE/Configuration or KDE/Settings
        kws = [KDE, Kw("X-KDE-settings")]
        kws += [k for m,
                k in _kdeCfgMap for x in parts if fnmatch.fnmatch(x, m)]
    else:
        kws = [k for m, k in _genMap for x in parts if fnmatch.fnmatch(x, m)]
    if len(kws):
        return KwS(*kws)
    else:
        return None


_kws = None


def scan():
    """Scan through all containers and return a list of found valid entries"""
    global _kws
    result = []
    msg("  desktop...", newline=False)
    for c in dirs:
        for w in os.walk(c):
            _kws = _dir2kws(w[0])
            if verbose > 1:
                msg("\nentering " + w[0])
            for x in fnmatch.filter(w[2], "*.desktop"):
                if verbose > 1:
                    msg("parsing %s..." % x, newline=False)
                try:
                    result.append(App(os.path.join(w[0], x)))
                    if verbose > 1:
                        msg("ok")
                except (NotDesktop, Prophet.NotSet) as e:
                    if verbose > 1:
                        msg("REJECTED : " + str(e))
                except UnicodeDecodeError:
                    if verbose > 1:
                        msg(".desktop UTF-8 decode issue in " + desktop)

    msg(" %d apps found" % len(result))
    return result

dirs = DesktopSet([os.path.join(p, x) for p in Prophet.prefixes for x in ("share/applications", "share/applnk", "share/gnome/apps")]) \
    + ("/usr/share/applications", "~/.kde/share/applnk",
       "~/.local/share/applications")

# Match section definition, e.g. [Global Settings]
_section = re.compile("\s*?\[\s*(.*)\s*\].*?")
# Match comments - lines starting with # or ;
_skip = re.compile("\s*[\#\;].*")
# Match any valid entry, e.g. key = value
_entry = re.compile("\s*?([^\=]*)\s*?\=\s*?(.*)\s*?")
# Match parametrized key, e.g. key[param]
_pkey = re.compile("(\w+)\s*\[\s*(\w+)\s*\]")

_DE = Kw("Desktop Entry")


def _parse(desktop):
    section = []
    inSection = False
    try:
        # .desktop files must be in UTF-8 encoding anyways.
        xs = codecs.open(desktop, "r", "utf-8").readlines()  # Py 2.4+
    except IOError:
        raise NotDesktop("i/o error while reading .desktop file")
    for x in xs:
        if re.match(_skip, x):
            continue
        rx = re.match(_section, x)
        if rx:
            inSection = (Kw(rx.group(1)) == _DE)
            continue
        if inSection:
            rx = re.match(_entry, x)
            if rx:
                section.append((rx.group(1), rx.group(2)))  # (key, value)

    if len(section):
        return _parseEntries(section)
    raise NotDesktop("[Desktop Entry] section is either absent or empty")


def _parseEntries(entries):
    e = {}
    for key, val in entries:
        rx = re.match(_pkey, key)
        if rx:
            pass  # TODO : _parseLocale
        else:
            e[key] = _parseValue(val)

    return e


def _parseValue(value):
    vl = [x.strip() for x in value.split(";") if not (x.strip() == "")]
    if len(vl) == 0:
        return None
    elif len(vl) > 1:
        return tuple(vl)
    else:
        return vl[0]


class NotDesktop(Exception):

    """Exception is raised when parser fails to decode entry"""


def _string(x):
    """Pick up the most appropriate string representation out of all specified variants
        Choice is made with respect to current locale and type of specified argument"""
    if isinstance(x, dict):
        return x[""]
    # FIXME : locale handling
    else:
        # Now argument is assumed to be a string
        return x


def _bool(x):
    """Convert string to boolean"""
    if x.lower() in ("yes", "true", "1"):
        return True
    else:
        return False


class App(Prophet.App):

    pref = 20

    def __new__(cls, desktop):
        self = object.__new__(cls)
        self.__setup__(desktop)
        return self

    def __setup__(self, desktop):
        self.desktop = _parse(desktop)
        try:
            if not (self.desktop["Type"] == "Application"):
                raise NotDesktop("Entry type is not Application")

        except KeyError:
            raise NotDesktop("Entry type is not specified")
        try:
            if _bool(self.desktop["NoDisplay"]):
                raise NotDesktop("NoDisplay is True")

        except KeyError:
            pass
        try:
            if _bool(self.desktop["Hidden"]):
                raise NotDesktop("Hidden is True")

        except KeyError:
            pass
        super(App, self).__setup__()
        del self.desktop

    def setName(self):
        for x in ("Name", "GenericName"):
            try:
                self.name = _string(self.desktop[x])
                return
            except KeyError:
                pass

        super(App, self).setName()

    def setComment(self):
        try:
            self.comment = _string(self.desktop["Comment"])
        except KeyError:
            super(App, self).setComment()

    def setKeywords(self):
        try:
            cats = self.desktop["Categories"]
            if type(cats) == tuple:
                self.keywords = KwS(*cats)
            else:
                self.keywords = KwS(cats)
        except KeyError:
            if _kws:
                self.keywords = _kws
            else:
                raise Prophet.NotSet("Categories not found")

        if fnmatch.fnmatch(self.exename, "*kcmshell"):
            self.keywords |= KwS(KDE, Kw("X-KDE-settings"))

    def setTerminal(self):
        try:
            self.terminal = _bool(self.desktop["Terminal"])
        except KeyError:
            super(App, self).setTerminal()

    def setExename(self):
        try:
            cmd = self.desktop["Exec"]
            # Currently all entry values are subject to semicolon splitting on parse:
            # /bin/sh -c "foo; bar"
            # While this is generally incorrect, it is as it is.
            # However, this causes the Exec processing to fail when command line contains semicolons
            # as a tuple is returned while string is expected.
            # As a workaround, we merge the tuples back for the Exec entries.
            # Reported by David Niklas.
            if isinstance(cmd, tuple):
                cmd = ";".join(map(str, cmd))
            # According to some reports, newer Debian-based distros
            # (notably, Ubuntu 5.10+) make an attempt to generate .desktop's
            # directly from their (Debian-style) database which means all those
            # ugly complex shell commands may emerge right here!
            # So we got to get rid of them for good.
            self.testGoodness(cmd)
            # Exec entry may contain the %-prefixed arguments that make no sense for us
            # and should be filtered out. If they're there, set exename to anything prior first
            # argument encountered
            scmd = cmd.split()
            if len(scmd) < 2:
                self.exename = cmd
            else:
                self.exename = scmd[0].strip()
                self.exeargs = ""
                for i in range(1, len(scmd)):
                    if "%" in scmd[i]:
                        break  # First %arg encountered terminates the list
                    try:
                        if scmd[i].startswith("-") and "%" in scmd[i + 1]:
                            break  # Filter out things like `-caption "%c"'

                    except IndexError:
                        pass
                    self.exeargs += scmd[i] + " "
                self.exeargs = self.exeargs.rstrip()
                if not len(self.exeargs):
                    del self.exeargs

        except KeyError:
            try:
                # As a last resort, check for TryExec entry. It should not
                # contain arguments so consume it all.
                self.exename = self.desktop["TryExec"]
            except KeyError:
                raise NotDesktop("Neither Exec nor TryExec specified")
                # If both entries are absent, this desktop file is totally
                # useless for us

    def setNativenv(self):
        super(App, self).setNativenv()
        # Not all Xfce apps have a distinguishing keywords therefore make an
        # additional check for the clues from its public name
        if fnmatch.fnmatch(self.name.lower(), "*xfce*"):
            self.nativenv = App.Xfce
        # A few GNOME apps do not manifest themselfs (according to their .desktop file)
        # as such
        elif fnmatch.fnmatch(os.path.basename(self.exename), "gnome*"):
            self.nativenv = App.GNOME



import os.path
import Prophet

from Prophet.Legacy import App as _App, ZeroG as _ZeroG, DropIn as _DropIn, ConsoleApp as _ConsoleApp, X11App as _X11App
from Keywords import Keyword as Kw, Set as KwS
from Prophet.Categories import *


class cforge(_App, _X11App):
    name = "CForge IDE"
    comment = "Code Forge IDE"
    keywords = KwS(Development, IDE)  # Qt, Motif ???


class ddd(_App, _X11App):
    name = "DDD debugger"
    comment = "X11 Display Data Debugger"
    keywords = KwS(Development, Debugger, Motif)


class eclipse(_DropIn, _X11App):
    name = "Eclipse"
    comment = "Eclipse Java IDE"
    keywords = KwS(Development, Debugger, Java, GTK)

    def relevant(self, path):
        if Prophet.isExe(os.path.join(path, "eclipse")):
            raise eclipse.StopDescention(path)


class eric3(_App, _X11App):
    name = "Eric IDE"
    comment = "Python IDE for X11"
    keywords = KwS(Development, IDE)


class estudio(_App, _X11App):
    name = "EiffelStudio"
    comment = "EiffelStudio IDE"
    keywords = KwS(Development, IDE, Debugger)

    def getPrefixes(self):
        return super(
            estudio,
            self).getPrefixes() + "$ISE_EIFFEL/studio/spec/$ISE_PLATFORM"


# TODO : Komodo

class jbuilder(_ZeroG, _X11App):
    name = "JBuilder"
    magic = ".*?Borland\sJBuilder.*?"
    comment = "Borland JBuilder IDE"
    keywords = KwS(Development, IDE, GUIDesigner, Java)


class fluid(_App, _X11App):
    name = "Fluid"
    comment = "FLTK interface builder"
    keywords = KwS(Development, IDE, GUIDesigner)


# FIXME : name's pretty generic, name clash possible
class fp(_App, _ConsoleApp):
    name = "FreePascal"
    comment = "Standart IDE for the FreePascal"
    keywords = KwS(Development, IDE, Core)


class gdb(_App, _ConsoleApp):
    name = "GDB"
    comment = "Standard GNU debugger"
    keywords = KwS(Development, Debugger, Core)


class idb(_App, _ConsoleApp):
    name = "IDB"
    comment = "Intel debugger"
    keywords = KwS(Development, Debugger)


class idea(_ZeroG, _X11App):
    exes = ["idea.sh", "idea"]
    name = "IDEA"
    magic = ".*?IntelliJ\sIDEA.*?"
    comment = "IntelliJ IDEA IDE"
    keywords = KwS(Development, IDE, GUIDesigner, Java)


class kylix(_App, _X11App):
    name = "Kylix"
    comment = "Borland Delphi IDE for UNIX"
    exes = ["startdelphi", "startkylix"]
    keywords = KwS(Development, IDE, GUIDesigner, Debugger)


class lazarus(_DropIn, _X11App):
    name = "Lazarus"
    comment = "Delphi-like FreePascal IDE"
    keywords = KwS(Development, Debugger, GUIDesigner, GTK)

    def relevant(self, path):
        if Prophet.isExe(os.path.join(path, "lazarus")):
            raise lazarus.StopDescention(path)


class motor(_App, _ConsoleApp):
    name = "Motor IDE"
    comment = "Console IDE with Borland look&feel"
    keywords = KwS(Development, IDE)


class netbeans(_DropIn, _X11App):
    name = "NetBeans"
    comment = "NetBeans Java IDE"
    keywords = KwS(Development, Debugger, Java, GTK)

    def relevant(self, path):
        if Prophet.isExe(os.path.join(path, "bin/netbeans")):
            raise netbeans.StopDescention(path)


class tkcvs(_App, _X11App):
    name = "TkCVS"
    comment = "X11 CVS frontend"
    keywords = KwS(Development, RevisionControl)


class ups(_App, _X11App):  # FIXME : generic name
    name = "UPS"
    comment = "Native X11 debugger"
    keywords = KwS(Development, Debugger)


class wpe(_App, _ConsoleApp):
    name = "WPE"
    comment = "A text-mode Borland-like IDE"
    keywords = KwS(Development, IDE)


class xwpe(_App, _X11App):
    name = "XWPE"
    comment = "Borland-like IDE for X11"
    keywords = KwS(Development, IDE)


class xxgdb(_App, _X11App):
    name = "XXGDB"
    comment = "X11 GDB frontend"
    keywords = KwS(Development, Debugger)



"""Module to handle keywords and sets of keywords"""

import types
import itertools


class Keyword(str):

    """Keyword is a case insensitive string with the spaces omitted,
        that is Keyword(Foo Bar) == Keyword(foobar)"""

    def __init__(self, s):
        t = ""
        if not s is None:
            for x in s.split():
                # FIXME : speed up the hashing by xor'ing the parts on which self is split
                # instead of performing this string concatenation
                t += x
        self.hash = hash(t.lower())
        str.__init__(self)

    def __hash__(self):
        return self.hash

    def __eq__(self, other):
        if not isinstance(other, Keyword):
            other = Keyword(other)
        # Use direct hash comparison to save a call to self.__hash__()
        return self.hash == other.hash


class Set(set):

    """Mutable set of keywords. Besides keywords, it can operate on ordinary strings as well,
        which get implicitly converted into the Keyword instances"""

    def __init__(self, *args):
        xargs = []
        for x in args:
            if isinstance(x, Keyword):
                xargs.append(x)
            else:
                xargs.append(Keyword(x))

        set.__init__(self, xargs)

    def __contains__(self, x):
        if isinstance(x, Keyword):
            return set.__contains__(self, x)
        else:
            return set.__contains__(self, Keyword(x))



from Prophet.Legacy import App as _App, ConsoleApp as _ConsoleApp, X11App as _X11App
from Keywords import Keyword as Kw, Set as KwS
from Prophet.Categories import *


class _Terminal(_App, _X11App):
    runFlag = "-e"

    def runCmd(self, cmd):
        return self._runCmd + cmd

    def setExecmd(self):
        super(_Terminal, self).setExecmd()
        self._runCmd = self.execmd + " " + self.runFlag + " "


class alacritty(_Terminal):
    name = "Alacritty"
    comment = "GPU-accelerated terminal emulator"
    keywords = KwS(TerminalEmulator)


class aterm(_Terminal):
    name = "Aterm"
    comment = "X terminal emulator"
    keywords = KwS(TerminalEmulator)


class c3270(_App, _ConsoleApp):
    name = "c3270"
    comment = "An IBM 3278/3279 terminal emulator"
    keywords = KwS(TerminalEmulator)


class dfm(_App, _X11App):  # FIXME : generic name
    name = "Dfm"
    comment = "OS/2 Workplace Shell clone"
    keywords = KwS(FileManager, Shell)


class emelfm(_App, _X11App):
    name = "emelFM"
    comment = "Simple file manager"
    keywords = KwS(FileManager, GTK)


class Eterm(_Terminal):
    name = "Eterm"
    comment = "Enlightenment terminal emulator"
    keywords = KwS(TerminalEmulator)


class fr(_App, _X11App):  # FIXME : generic name
    name = "FileRunner"
    comment = "File manager written in Tcl/Tk"
    keywords = KwS(FileManager, Shell)


class git(_App, _ConsoleApp):
    name = "GIT"
    comment = "GNU interactive tools"
    keywords = KwS(FileManager, Shell, Utility)


class gmc(_App, _X11App):
    name = "GnomeMC"
    comment = "GNOME version of Midnight Commander"
    keywords = KwS(FileManager, Shell, GNOME)


class _gterm(_Terminal):
    exes = ["gnome2-terminal", "gnome-terminal", "gterminal"]
    name = "GNOME terminal"
    comment = "X terminal emulator for GNOME"
    keywords = KwS(TerminalEmulator, GNOME)


class _konsole(_Terminal):
    exes = ["konsole"]
    name = "Konsole"
    comment = "X terminal emulator for KDE"
    keywords = KwS(TerminalEmulator, KDE)


class kterm(_Terminal):
    name = "Kterm"
    comment = "Japanese X terminal emulator"
    keywords = KwS(TerminalEmulator)


class mc(_App, _ConsoleApp):
    name = "MC"
    comment = "Midnight Commander"
    exes = ["midc", "mc"]
    keywords = KwS(FileManager, Shell)


class qterminal(_Terminal):
    name = "QTerminal"
    comment = "Lightweight Qt-based terminal emulator"
    keywords = KwS(TerminalEmulator)


class rdesktop(_App, _X11App):
    name = "Rdesktop"
    comment = "MS Windows terminal client"
    keywords = KwS(TerminalEmulator)


class rox(_App, _X11App):
    name = "ROX"
    comment = "RISCOS-like file manager"
    keywords = KwS(FileManager, Shell)


class rxvt(_Terminal):
    name = "Rxvt"
    comment = "Our X terminal emulator"
    keywords = KwS(TerminalEmulator)


class sakura(_Terminal):
    name = "Sakura"
    comment = "Terminal emulator using GTK and VTE"
    keywords = KwS(TerminalEmulator)


class st(_Terminal):
    name = "ST"
    comment = "Simple suckless terminal emulator"
    keywords = KwS(TerminalEmulator)


class tuxcmd(_App, _X11App):
    name = "TuxCommander"
    comment = "File manager for X"
    keywords = KwS(FileManager, Shell)


class wterm(_Terminal):
    name = "Wterm"
    comment = "X terminal emulator"
    keywords = KwS(TerminalEmulator)


class x3270(_App, _X11App):
    name = "x3270"
    comment = "An IBM 3278/3279 terminal emulator"
    keywords = KwS(TerminalEmulator)


class xfiles(_App, _X11App):
    name = "X-Files"
    comment = "File manager for X"
    exes = ["X-Files"]
    keywords = KwS(FileManager, Shell)


class _xfterm(_Terminal):
    exes = ["xfce4-terminal", "xfterm4", "xfterm"]
    name = "Xfterm"
    comment = "X terminal emulator for Xfce"
    keywords = KwS(TerminalEmulator, Kw("Xfce"))


class xnc(_App, _X11App):
    name = "XNC"
    comment = "X Nothern Captain - NC clone"
    keywords = KwS(FileManager, Shell)


class xplore(_App, _X11App):
    name = "Xplorer"
    comment = "X file manager"
    keywords = KwS(FileManager, Shell, Motif)


class xterm(_Terminal):
    name = "Xterm"
    comment = "X terminal emulator"
    keywords = KwS(TerminalEmulator, Core)


class xvt(_Terminal):
    name = "Xvt"
    comment = "X terminal emulator"
    keywords = KwS(TerminalEmulator)


class xwc(_App, _X11App):
    name = "XWinCommander"
    comment = "X file manager"
    keywords = KwS(FileManager, Shell)



import MenuMaker

from MenuMaker import indent, writeFullMenu

menuFile = "~/.config/openbox/menu.xml"


def _map(x):
    for d, s in (("&amp;", "&"), ("\'", "\"")):
        x = x.replace(s, d)
    return x


class Sep(object):

    def emit(self, level):
        return ['%s<separator/>' % indent(level)]


class App(object):

    def emit(self, level):
        x = indent(level)
        xx = indent(level + 1)
        cmd = self.app.execmd
        if self.app.terminal:
            cmd = MenuMaker.terminal.runCmd(cmd)
        return [
            '%s<item label="%s"> <action name="Execute">' % (
                x, _map(self.app.name)),
            '%s<execute>%s</execute>' % (xx, cmd),
            '%s</action> </item>' % x
        ]


class Menu(object):
    id = 0

    def __init__(self):
        super(Menu, self).__init__()
        self.id = Menu.id
        Menu.id += 1

    def emit(self, level):
        menu = ['%s<menu id="%s" label="%s">' %
                (indent(level), self.id, _map(self.name))]
        for x in self:
            menu += x.emit(level + 1)
        menu.append('%s</menu>' % indent(level))
        return menu


class Root(object):
    name = "OpenBox 3"

    def __init__(self, subs):
        if writeFullMenu:
            subs += [MenuMaker.Sep(), SysMenu()]
        super(Root, self).__init__(subs)
        self.id = "root-menu"

    def emit(self, level):
        if writeFullMenu:
            menu = [
                '<?xml version="1.0" encoding="UTF-8"?>',
                '<openbox_menu>'
            ]
            menu += super(Root, self).emit(level + 1)
            menu.append('</openbox_menu>')
            return menu
        else:
            menu = []
            for x in self:
                menu += x.emit(level)
            return menu


class SysMenu(MenuMaker.Menu):
    name = "OpenBox"

    def __init__(self):
        subs = [
            X('<menu id="client-list-menu"/>'),
            X('<item label="Reconfigure"> <action name="Reconfigure"/> </item>'),
            MenuMaker.Sep(),
            X('<item label="Exit"> <action name="Exit"/> </item>')]
        super(SysMenu, self).__init__(subs)
        self.align = MenuMaker.Entry.StickBottom


class X(MenuMaker.Entry):

    def __init__(self, x):
        super(X, self).__init__()
        self.align = MenuMaker.Entry.StickBottom
        self.x = x

    def emit(self, level):
        return [indent(level) + self.x]



import os.path
import MenuMaker
import Prophet
import Prophet.Legacy

from MenuMaker import indent, writeFullMenu
from MenuMaker.xBox import X, Sep, App, Menu, Root as _Root

menuFile = "~/.fluxbox/menu"


class fluxbox(Prophet.Legacy.App):
    pass


class Root(_Root):
    name = "FluxBox"

    def __init__(self, subs):
        if writeFullMenu:
            subs += [MenuMaker.Sep(), SysMenu()]
        super(Root, self).__init__(subs)


class StylesMenu(MenuMaker.Menu):
    name = "Styles"

    def __init__(self, dirs):
        super(StylesMenu, self).__init__(
            [X(x, "stylesdir") for x in dirs if os.path.isdir(os.path.expanduser(x))])
        self.align = MenuMaker.Entry.StickBottom


class SysMenu(MenuMaker.Menu):
    name = "FluxBox"

    def __init__(self):
        subs = [X("Workspaces", "workspaces")]
        try:
            subs += [StylesMenu([os.path.join(fluxbox().prefix,
                                              "share/fluxbox/styles"),
                                 "~/.fluxbox/styles"])]
        except Prophet.NotSet:
            pass
        subs += [
            X("Configure", "config"),
            X("Reconfig", "reconfig"),
            X("Restart", "restart"),
            MenuMaker.Sep(),
            X("Exit", "exit")
        ]
        super(SysMenu, self).__init__(subs)
        self.align = MenuMaker.Entry.StickBottom



import MenuMaker
import Prophet.Legacy

from MenuMaker import indent, writeFullMenu

menuFile = "~/.jwmrc"


def _map(x):
    for d, s in (("&amp;", "&"), ("\'", "\"")):
        x = x.replace(s, d)
    return x


class Sep(object):

    def emit(self, level):
        return ['%s<Separator/>' % indent(level)]


class App(object):

    def emit(self, level):
        cmd = self.app.execmd
        if self.app.terminal:
            cmd = MenuMaker.terminal.runCmd(cmd)
        return [
            '%s<Program label="%s">%s</Program>' %
            (indent(level),
                _map(self.app.name),
                cmd
            )
        ]


class Menu(object):

    def emit(self, level):
        result = ['%s<Menu icon="folder.png" label="%s">' % (indent(level), _map(self.name))]
        for x in self:
            result += x.emit(level + 1)
        result.append('%s</Menu>' % indent(level))
        return result


class Root(object):

    def __init__(self, subs):
        super(Root, self).__init__(subs)

    def emit(self, level):
        if writeFullMenu:
            menu = [
                '<?xml version="1.0"?>',
                '<JWM>',
            ]
            menu += self.emitMenu(level)
            menu.append(defaultConfig)
            menu.append('</JWM>')
            return menu
        else:
            return self.emitMenu(level)

    def emitMenu(self, level):
        menu = ['<RootMenu onroot="12">']
        for x in self:
            menu += x.emit(level)
        sysmenu = [
            MenuMaker.Sep(),
            X('<Restart label="Restart" icon="restart.png"/>'),
            X('<Exit label="Exit" confirm="true" icon="quit.png"/>')
        ]
        for x in sysmenu:
            menu += x.emit(level)
        menu += ['</RootMenu>']
        return menu



class X(MenuMaker.Entry):

    def __init__(self, x):
        super(X, self).__init__()
        self.align = MenuMaker.Entry.StickBottom
        self.x = x

    def emit(self, level):
        return [indent(level) + self.x]


# Excerpt from /etc/system.jwmrc as of 2.3.7
defaultConfig = '''
    <!-- Options for program groups. -->
    <Group>
        <Option>tiled</Option>
        <Option>aerosnap</Option>
    </Group>
    <Group>
        <Class>Pidgin</Class>
        <Option>sticky</Option>
    </Group>
    <Group>
        <Name>xterm</Name>
        <Option>vmax</Option>
    </Group>
    <Group>
        <Name>xclock</Name>
        <Option>drag</Option>
        <Option>notitle</Option>
    </Group>

    <!-- Tray at the bottom. -->
    <Tray x="0" y="-1" autohide="off">

        <TrayButton icon="jwm-blue">root:1</TrayButton>
        <Spacer width="2"/>
        <TrayButton label="_">showdesktop</TrayButton>
        <Spacer width="2"/>

        <Pager labeled="true"/>

        <TaskList maxwidth="256"/>

        <Dock/>
        <Clock format="%H:%M"><Button mask="123">exec:xclock</Button></Clock>

    </Tray>

    <!-- Visual Styles -->
    <WindowStyle>
        <Font>Sans-9:bold</Font>
        <Width>4</Width>
        <Height>21</Height>
        <Corner>3</Corner>
        <Foreground>#FFFFFF</Foreground>
        <Background>#555555</Background>
        <Outline>#000000</Outline>
        <Opacity>0.5</Opacity>
        <Active>
            <Foreground>#FFFFFF</Foreground>
            <Background>#0077CC</Background>
            <Outline>#000000</Outline>
            <Opacity>1.0</Opacity>
        </Active>
    </WindowStyle>
    <TrayStyle group="true" list="all">
        <Font>Sans-9</Font>
        <Background>#333333</Background>
        <Foreground>#FFFFFF</Foreground>
        <Outline>#000000</Outline>
        <Opacity>0.75</Opacity>
    </TrayStyle>
    <TaskListStyle>
      <Font>Sans-9</Font>
      <Active>
        <Foreground>#FFFFFF</Foreground>
        <Background>#555555</Background>
      </Active>
      <Foreground>#FFFFFF</Foreground>
      <Background>#333333</Background>
    </TaskListStyle>
    <PagerStyle>
        <Outline>#000000</Outline>
        <Foreground>#555555</Foreground>
        <Background>#333333</Background>
        <Text>#FFFFFF</Text>
        <Active>
            <Foreground>#0077CC</Foreground>
            <Background>#004488</Background>
        </Active>
    </PagerStyle>
    <MenuStyle>
        <Font>Sans-9</Font>
        <Foreground>#FFFFFF</Foreground>
        <Background>#333333</Background>
        <Outline>#000000</Outline>
        <Active>
            <Foreground>#FFFFFF</Foreground>
            <Background>#0077CC</Background>
        </Active>
        <Opacity>0.85</Opacity>
    </MenuStyle>
    <PopupStyle>
        <Font>Sans-9</Font>
        <Foreground>#000000</Foreground>
        <Background>#999999</Background>
    </PopupStyle>

    <!-- Path where icons can be found.
         IconPath can be listed multiple times to allow searching
         for icons in multiple paths.
      -->
    <IconPath>
        /usr/share/icons/wm-icons/32x32-aquafusion
    </IconPath>
    <IconPath>
        /usr/local/share/jwm
    </IconPath>

    <!-- Virtual Desktops -->
    <!-- Desktop tags can be contained within Desktops for desktop names. -->
    <Desktops width="4" height="1">
        <!-- Default background. Note that a Background tag can be
              contained within a Desktop tag to give a specific background
              for that desktop.
         -->
        <Background type="solid">#111111</Background>
    </Desktops>

    <!-- Double click speed (in milliseconds) -->
    <DoubleClickSpeed>400</DoubleClickSpeed>

    <!-- Double click delta (in pixels) -->
    <DoubleClickDelta>2</DoubleClickDelta>

    <!-- The focus model (sloppy or click) -->
    <FocusModel>sloppy</FocusModel>

    <!-- The snap mode (none, screen, or border) -->
    <SnapMode distance="10">border</SnapMode>

    <!-- The move mode (outline or opaque) -->
    <MoveMode>opaque</MoveMode>

    <!-- The resize mode (outline or opaque) -->
    <ResizeMode>opaque</ResizeMode>

    <!-- Key bindings -->
    <Key key="Up">up</Key>
    <Key key="Down">down</Key>
    <Key key="Right">right</Key>
    <Key key="Left">left</Key>
    <Key key="h">left</Key>
    <Key key="j">down</Key>
    <Key key="k">up</Key>
    <Key key="l">right</Key>
    <Key key="Return">select</Key>
    <Key key="Escape">escape</Key>

    <Key mask="A" key="Tab">nextstacked</Key>
    <Key mask="A" key="F4">close</Key>
    <Key mask="A" key="#">desktop#</Key>
    <Key mask="A" key="F1">root:1</Key>
    <Key mask="A" key="F2">window</Key>
    <Key mask="A" key="F10">maximize</Key>
    <Key mask="A" key="Right">rdesktop</Key>
    <Key mask="A" key="Left">ldesktop</Key>
    <Key mask="A" key="Up">udesktop</Key>
    <Key mask="A" key="Down">ddesktop</Key>
'''


from Prophet.Legacy import App as _App, ConsoleApp as _ConsoleApp, X11App as _X11App
from Keywords import Keyword as Kw, Set as KwS
from Prophet.Categories import *


class xpcd(_App, _X11App):
    name = "XPCD"
    comment = "PhotoCD Viewer"
    keywords = KwS(Graphics, Photograph, Viewer)


class krita(_App, _X11App):
    name = "Krita KDE Bitmap Graphics Editor"
    comment = "Krita KDE Bitmap graphics editor"
    keywords = KwS(Graphics, RasterGraphics, KDE)


class ufraw(_App, _X11App):
    name = "Unidentified Flying RAW"
    comment = "Unidentified Flying RAW"
    keywords = KwS(Graphics, Photograph)
